import os
import time
import shutil
import random
import logging
import cStringIO
from xml.dom import minidom
from xml.sax import saxutils
from xml.parsers.expat import ExpatError

import osutil
import subproc
from encoder import EncoderFile

import pmm_config
import migration_handler_config
from pmm_api_xml_protocols import AgentCredentials, SelectedObjects, MigrationTaskInfoMigratingObjects, \
MigrationObjectList, IPMapping, ConflictResolutionRules, OwnersMapping, DBMapping, TaskStatus, \
MigrationTaskInfo, Session, SessionSteps, SessionStep, SessionList, Finished, ScoutResult, DumpTransformerOptions


_logger = logging.getLogger("pmm.migration_handler.session")


class MigrationSessionException(Exception):
    def __init__(self, message, migration_session_id):
        Exception.__init__(self, message)
        self.__migration_session_id = migration_session_id
        _logger.info("MigrationSessionException generated: " + str(self.__migration_session_id))

    def get_migration_session_id(self):
        return self.__migration_session_id


class MigrationSession:
    def __init__(self, session_id):
        self.__session_id = session_id
        self.__session_path = os.path.join( migration_handler_config.migration_session_dir(), self.__session_id)
        self.__migration_pid_file = 'migration_pid'
        self.__migration_pid = os.path.join( self.__session_path, self.__migration_pid_file)
        self.__worker_pid_file = 'migration_worker_pid'
        self.__worker_pid = os.path.join( self.__session_path, self.__worker_pid_file)
        self.__migration_result = os.path.join( self.__session_path, 'migration.result')
        self.__last_time = os.path.join( self.__session_path, 'last_time')
        if not os.path.isdir(self.__session_path):
            self.__raiseException("Could not initialize session directory")
        
        # self.__objects contains a dictionary:
        # Key - item is a name of the file to store corresponding object
        # Value - is tuple (object_required, object_type, object_element_name), where
        #    object_required - is object required for session or not.
        #        'getObject' will raise MigrationSessionException when loading failed for required object.
        #    object_type - is a object type class (imported from pmm-api_xml_protocols).
        #    object_element_name - is an xml element name, used for storing object string as a object file content.
        self.__objects = {'agent_credentials': (True, AgentCredentials,'agent-credentials'),
                          'scout_result': (False, ScoutResult,'scout-result'),
                          'objects_list': (False, MigrationObjectList,'migration-objects-list'),
                          'ip_mapping': (False, IPMapping,'ip-mapping'),
                          'ip_for_mapping': (False, IPMapping,'ip-mapping'),
                          'restore_policy': (False, ConflictResolutionRules,'conflict-resolution-rules'),
                          'selected_objects': (False, MigrationObjectList,'migration-objects-list'),
                          'dump_transformer_options': (False, DumpTransformerOptions,'dump-transformer-options'),
                          'owner_mapping': (False, OwnersMapping,'owner-mapping'),
                          'db_mapping': (False, DBMapping,'db-mapping'),
                          'restore_task_status': (False, TaskStatus,'restore'),
                          'backup_finished_status': (False, TaskStatus,'status'),
                          'backup_current_status': (False, TaskStatus,'status'),
                          'additional_info': (False, MigrationTaskInfo,'additional-info')}
        # self.__values contains a list: 
        # Key is a name of the file to store corresponding value.
        # Value - is value_required, where
        #     value_required - is value required for session or not.
        self.__values = {'start_time': True,
                         'agent_name': False,
                         'agent_dir': False,
                         'owner_guid': False,
                         'dump-name': False,
                         'migration_task_id': False,
                         'dst_temporary_directory': False,
                         'source_os': False,
                         'source_host': False,
                         'step': False,
                         'result': False,
                         'last_time': False,
                         'failed-restore.result': False}
        
        # self.__flags contains a list: 
        # List item is a name of the file to store corresponding flag.
        #     If file does not exist - flag is False, otherwise is True
        self.__flags = ['agent_uploaded', 'migration_task_started', 'stopped']
        
        # check for required data objects
        if self.getObject('agent_credentials') is None:
            # Now always is not None due to current AgentCredentials implementation
            self.__raiseException("Wrong format for agent credentials")
        
        # check for required data values
        self.getSavedTime('start_time')

    # create new dir for migration session
    def createSession(session_id):
        session_path = os.path.join( migration_handler_config.migration_session_dir(), session_id)
        start_time_storage_filename = os.path.join( session_path, 'start_time')
        time_str = time.strftime('%Y%m%d%H%M%S',time.localtime())
        try:
            start_time_storage = open(start_time_storage_filename, 'wt')
            try:
                start_time_storage.write(time_str)
            finally:
                start_time_storage.close()
        except IOError, x:
            errmsg = "IOError raised in createSession when try to save start time file '%s': %s" % (start_time_storage_filename, str(x))
            _logger.error(errmsg)
            raise MigrationSessionException(errmsg, session_id)
        
        session = MigrationSession(session_id)
        return session

    createSession = staticmethod(createSession)

    def __raiseException(self, message):
        raise MigrationSessionException(message, self.__session_id)

    def getSessionPath(self):
        return self.__session_path

    def getSessionId(self):
        return self.__session_id

    def setObject(self, name_, object_):
        if name_ not in self.__objects:
            self.__raiseException("Invalid data object name '%s' for migration session" % name_)
        object_type, object_element_name = self.__objects[name_][1:]
        if not isinstance(object_, object_type):
            self.__raiseException("Invalid object type '%s' for migration session object '%s'" % (object_type, name_))
        object_storage_filename = os.path.join(self.__session_path, name_)
        object_storage = open(object_storage_filename, 'wt')
        try:
            resp_str = cStringIO.StringIO()
            resp_encoded = EncoderFile(resp_str, "utf-8")
            resp_encoded.write('<?xml version="1.0" encoding="UTF-8"?>\n')
            object_.export(resp_encoded, 0, name_ = object_element_name)
            object_storage.write(resp_str.getvalue())
        finally:
            object_storage.close()

    def getObject(self, name_):
        if name_ not in self.__objects:
            self.__raiseException("Invalid data object name '%s' for migration session" % name_)
        object_required, object_type = self.__objects[name_][:-1]
        object_ = None
        object_storage_filename = os.path.join(self.__session_path, name_)
        if not os.path.isfile(object_storage_filename):
            _logger.warning("Session data object file '%s' does not exist" % object_storage_filename)
            if object_required:
                self.__raiseException("Required session data object '%s' is not found in migration session" % name_)
        else:
            try:
                object_ = object_type.factory()
                object_.build(minidom.parse(object_storage_filename).childNodes[0] )
            except IOError, x:
                _logger.warning("IOError raised in getObject('%s'): %s" % (name_, str(x)))
            except ExpatError, x:
                _logger.warning("ExpatError raised in getObject('%s'): %s" % (name_, str(x)))
        return object_

    def setValue(self, name_, value_):
        if name_ not in self.__values:
            self.__raiseException("Invalid data value name '%s' for migration session" % name_)
        value_storage_filename = os.path.join(self.__session_path, name_)
        value_storage = open(value_storage_filename,"wt")
        try:
            value_storage.write(str(value_))
        finally:
            value_storage.close()

    def getValue(self, name_):
        if name_ not in self.__values:
            self.__raiseException("Invalid data value name '%s' for migration session" % name_)
        value_required = self.__values[name_]
        value_ = None
        value_storage_filename = os.path.join(self.__session_path, name_)
        if not os.path.isfile(value_storage_filename):
            _logger.warning("Session data value file '%s' does not exist" % value_storage_filename)
            if value_required:
                self.__raiseException("Required session data value '%s' is not found in migration session" % name_)
        else:
            value_storage = open(value_storage_filename,"rt")
            try:
                try:
                    value_ = value_storage.read()
                except IOError, x:
                    _logger.warning("IOError raised in getValue('%s'): %s" % (name_, str(x)))
            finally:
                value_storage.close()
        return value_

    def setFlag(self, name_, value_ = True):
        if name_ not in self.__flags:
            self.__raiseException("Invalid data flag name '%s' for migration session" % name_)
        flag_storage_filename = os.path.join(self.__session_path, name_)
        if not value_:
            if os.path.isfile(flag_storage_filename):
                try:
                    os.unlink(flag_storage_filename)
                except OSError, x:
                    _logger.warning("OSError raised in setFlag when try to remove flag file '%s': %s" % (flag_storage_filename, str(x)))
        else:
            try:
                flag_storage = open(flag_storage_filename,"wt")
                try:
                    flag_storage.write('True')
                finally:
                    flag_storage.close()
            except IOError, x:
                _logger.warning("IOError raised in setFlag when try to create flag file '%s': %s" % (flag_storage_filename, str(x)))

    def getFlag(self, name_):
        if name_ not in self.__flags:
            self.__raiseException("Invalid data flag name '%s' for migration session" % name_)
        flag_storage_filename = os.path.join(self.__session_path, name_)
        return os.path.isfile(flag_storage_filename)

    def setRestoreFailed(self, result_value):
        self.setValue('failed-restore.result', result_value)
        finished_status_error_object = Finished( status = 'error' , log_location = os.path.join(self.__session_path, 'failed-restore.result') )
        task_status_object = TaskStatus( finished = finished_status_error_object )
        self.setObject('restore_task_status', task_status_object)

    def getStepsPassed(self):
        steps_passed_object = SessionSteps()
        if self.getFlag('agent_uploaded'):
            session_step_object = SessionStep( id = 'agent-uploaded' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('objects_list') is not None:
            session_step_object = SessionStep( id = 'objects-list-received' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('selected_objects') is not None:
            session_step_object = SessionStep( id = 'migrated-objects-selected' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('ip_mapping') is not None:
            session_step_object = SessionStep( id = 'ip-mapping-set' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('owner_mapping') is not None:
            session_step_object = SessionStep( id = 'owners-set' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('db_mapping') is not None:
            session_step_object = SessionStep( id = 'db-mapping-set' )
            steps_passed_object.add_step(session_step_object)
        if self.getObject('restore_policy') is not None:
            session_step_object = SessionStep( id = 'restore-policy-set' )
            steps_passed_object.add_step(session_step_object)
        if self.getFlag('migration_task_started'):
            session_step_object = SessionStep( id = 'migration-task-started' )
            steps_passed_object.add_step(session_step_object)
        if len(steps_passed_object.get_step()) == 0:
            return None
        else:
            return steps_passed_object

    #
    # Begin Worker process control methods

    def setProcess(self, subprocess):
        pid = -1
        if subprocess is not None:
            pid = subprocess.get_pid()
        pid_str = str(pid)
        _logger.debug("add '%s' to session worker processes" % pid_str)
        pid_file_name = self.__worker_pid + '.' + pid_str
        pid_file = open(pid_file_name,'w')
        try:
            pid_file.write(' ')
        finally:
            pid_file.close()

    def unsetProcess(self, subprocess):
        pid_str = str(subprocess.get_pid())
        _logger.debug("remove '%s' from session worker processes" % pid_str)
        pid_file_name = self.__worker_pid + '.' + pid_str
        try:
            os.unlink(pid_file_name)
        except OSError, x:
            _logger.warning("OSError raised in unsetProcess when try to remove pid file %s: %s" % (pid_file_name, str(x)))

    def setMigrationProcess(self, pid):
        pid_str = str(pid)
        _logger.debug("set '%s' as session migration process" % pid_str)
        pid_file_name = self.__migration_pid + '.' + pid_str
        pid_file = open(pid_file_name,'w')
        try:
            pid_file.write(' ')
        finally:
            pid_file.close()

    def getMigrationPid(self):
        prefix = self.__migration_pid_file + '.'
        for dir_item in os.listdir(self.__session_path):
            if dir_item.startswith(prefix) and os.path.isfile(os.path.join(self.__session_path, dir_item)):
                try:
                    pid_str = dir_item[len(prefix):]
                    pid = int(pid_str)
                    return pid
                except:
                    pass
        return None

    def createWorker(self, filename, args = [], vars = os.environ, stdin = "", stdout = None, stderr = None):
        return subproc.CmdLine(filename, self, args, vars, stdin, stdout, stderr)

    def runWorker(self, cmd):
        if cmd.process_handler != self:
            _logger.error("Trying run worker created with another migration session instance")
            raise Exception("Trying run worker created with another migration session instance")
        cmd.spawn()

    def getWorkerPids(self):
        ret = []
        prefix = self.__worker_pid_file + '.'
        for dir_item in os.listdir(self.__session_path):
            if dir_item.startswith(prefix) and os.path.isfile(os.path.join(self.__session_path, dir_item)):
                try:
                    pid_str = dir_item[len(prefix):]
                    pid = int(pid_str)
                    ret.append(pid)
                except:
                    pass
        return ret

    # End Worker process control methods

    def __setBackupStatusImpl(self, backup_object_name, backup_status):
        if backup_status is None:
            backup_storage_filename = os.path.join(self.__session_path, backup_object_name)
            if os.path.isfile(backup_storage_filename):
                try:
                    os.unlink(backup_storage_filename)
                except OSError, x:
                    _logger.warning("OSError raised in __setBackupStatusImpl when try to remove backup status file '%s': %s" % (backup_storage_filename, str(x)))
        else:
            self.setObject(backup_object_name, backup_status)

    def setBackupFinishedStatus(self, backup_status):
        self.__setBackupStatusImpl('backup_finished_status', backup_status)

    def getBackupFinishedStatus(self):
        return self.getObject('backup_finished_status')

    def setBackupCurrentStatus(self, backup_status):
        self.__setBackupStatusImpl('backup_current_status', backup_status)

    def getBackupCurrentStatus(self):
        return self.getObject('backup_current_status')

    def getMigrationStatus(self):
        if not os.path.isfile(self.__migration_result):
            _logger.info("Migration status file not found in session")
            return None
        
        doc = minidom.parse(self.__migration_result)
        return self.__walkMigrationResult(doc)

    def __walkMigrationResult(self, node):
        if isinstance(node, minidom.Element) and node.localName == 'execution-result':
            return node.getAttribute('status')
        for child in node.childNodes:
            ret = self.__walkMigrationResult(child)
            if ret != None:
                return ret
        return None

    def setTime(self, time_data_value):
        time_str = time.strftime('%Y%m%d%H%M%S',time.localtime())
        self.setValue(time_data_value, time_str)

    def getSavedTime(self, time_data_name):
        time_value = self.getValue(time_data_name)
        time_value_time = None
        if time_value is not None:
            try:
                time_value_time = time.strptime(time_value, '%Y%m%d%H%M%S') 
            except:
                pass
        if time_value_time is None:
            # raise exception for required session data value
            if self.__values[time_data_name]:
                self.__raiseException("Wrong format for start time")
            else:
                return time.localtime()
        else:
            return time_value_time

    def getWorkTime(self):
        return time.mktime(self.getSavedTime('last_time')) - time.mktime(self.getSavedTime('start_time'))

    def getTransferedData(self):
        return osutil.getDirectorySize(self.__session_path)

    def getMigratedDomains(self):
        ret = 0
        try:
            selectedObjects = self.getObject('selected_objects')
            if selectedObjects:
                if selectedObjects.get_reseller():
                    for reseller in  selectedObjects.get_reseller():
                        if reseller.get_domains() and reseller.get_domains().get_domain(): 
                            ret += len( reseller.get_domains().get_domain() )
                        if reseller.get_clients() and reseller.get_clients().get_client():
                            for client in  reseller.get_clients().get_client():
                                if client.get_domains() and client.get_domains().get_domain():
                                    ret += len( client.get_domains().get_domain() )
                if selectedObjects.get_client():
                    for client in  selectedObjects.get_client():
                        if client.get_domains() and client.get_domains().get_domain():
                            ret += len( client.get_domains().get_domain() )
                if selectedObjects.get_domain():
                    ret += len( selectedObjects.get_domain() )
        except Exception, x:
            _logger.warning("getMigratedDomains raised exception: %s" % str(x))
        return ret

    def quoteXml(self, inStr):
        s1 = ( isinstance(inStr, basestring) and inStr or '%s' % inStr)
        return saxutils.escape(s1)

    def PostMigration(self):
        post_filename = os.path.join( self.__session_path, "migration-data" )
        if os.path.isfile(post_filename):
            _logger.warning("'migration-data' file already exists in migration session.")
            return
        cmd = subproc.AsyncCmdLine(pmm_config.plesk_runner())
        for plesk_runner_arg in pmm_config.plesk_runner_args():
            cmd.arg(plesk_runner_arg)
        cmd.arg( migration_handler_config.post_migration() )
        cmd.arg( "post" )
        cmd.arg( post_filename )
        data = "<migration-data>"
        data += "<os>" + self.quoteXml( self.getValue('source_os')) + "</os>"
        data += "<host>" + self.quoteXml( self.getValue('source_host') ) + "</host>"
        data += "<platform>"  + self.quoteXml( self.getValue('agent_name') ) + "</platform>"
        data += "<domains>" + self.quoteXml( self.getMigratedDomains() ) + "</domains>"
        data += "<time>" + self.quoteXml( self.getWorkTime() ) + "</time>"
        data += "<size>" + self.quoteXml( self.getTransferedData() ) + "</size>"
        data += "<step>" + self.quoteXml( self.getValue('step') ) + "</step>"
        data += "<result>" + self.quoteXml( self.getValue('result') ) + "</result>"
        data += "</migration-data>"
        post_file = open(post_filename, 'wt')
        try:
            try:
                post_file.write(data)
            except IOError, x:
                _logger.warning("IOError raised in PostMigration: %s" % str(x))
        finally:
            post_file.close()
        cmd.asyncSpawn()

    def getSelectedObjectsPath(self):
        return os.path.join( self.__session_path, 'selected_objects')

    def getAdditionalInfo(self):
        additional_info = self.getObject('additional_info')
        if additional_info is not None:
            return additional_info
        
        source_hostname_str = ''
        agent_credentials = self.getObject('agent_credentials')
        host_credentials = agent_credentials.get_host_credentials()
        if host_credentials is not None:
            source_hostname_str = host_credentials.get_hostname()
        migrating_objects_object = None
        
        migration_objects_list = self.getObject('selected_objects')
        if migration_objects_list is not None:
            # start count objects otherwise
            resellers_count = 0
            clients_count = 0
            domains_count = 0
            selected_objects_list = []
            resellers = migration_objects_list.get_reseller()
            if resellers is not None:
                for reseller in resellers:
                    resellers_count += 1
                    
                    clients = reseller.get_clients()
                    if clients is not None:
                        for client in clients.get_client():
                            clients_count += 1
                            
                            domains = client.get_domains()
                            if domains is not None:
                                for domain in domains.get_domain():
                                    domains_count += 1
                    domains = reseller.get_domains()
                    if domains is not None:
                        for domain in domains.get_domain():
                            domains_count += 1
            
            clients = migration_objects_list.get_client()
            if clients is not None:
                for client in clients:
                    clients_count += 1
                    
                    domains = client.get_domains()
                    if domains is not None:
                        for domain in domains.get_domain():
                            domains_count += 1
            
            domains = migration_objects_list.get_domain()
            if domains is not None:
                for domain in domains:
                    domains_count += 1
            
            if resellers_count > 0:
                selected_objects_object = SelectedObjects.factory()
                selected_objects_object.set_type('resellers')
                selected_objects_object.set_amount(resellers_count)
                selected_objects_list.append(selected_objects_object)
            
            if clients_count > 0:
                selected_objects_object = SelectedObjects.factory()
                selected_objects_object.set_type('clients')
                selected_objects_object.set_amount(clients_count)
                selected_objects_list.append(selected_objects_object)
            
            if domains_count > 0:
                selected_objects_object = SelectedObjects.factory()
                selected_objects_object.set_type('domains')
                selected_objects_object.set_amount(domains_count)
                selected_objects_list.append(selected_objects_object)
            
            migrating_objects_object = MigrationTaskInfoMigratingObjects( selected_objects = selected_objects_list )
        
        additional_info = MigrationTaskInfo( source_hostname = source_hostname_str, migrating_objects = migrating_objects_object )
        self.setObject('additional_info',additional_info)
        return additional_info


migrationSessionManager = None


def getMigrationSessionManager():
    global migrationSessionManager
    if not migrationSessionManager:
        migrationSessionManager = MigrationSessionManager()
    return migrationSessionManager


class MigrationSessionManager:
    def __init__(self):
        self.__session_ids = None

    def __get_session(self, session):
        migration_session = None
        if isinstance(session, str) or isinstance(session, unicode):
            try:
                migration_session = MigrationSession(session)
            except MigrationSessionException, x:
                _logger.warning("Exception raised while trying to initialize migration session '%s': %s" % (session, str(x)))
        elif isinstance(session, MigrationSession):
            migration_session = session
        return migration_session

    def __get_session_ids(self):
        self.__session_ids = []
        migration_session_dir = migration_handler_config.migration_session_dir()
        if os.path.isdir(migration_session_dir):
            names = os.listdir(migration_session_dir)
            for name in names:
                fullname = os.path.join(migration_session_dir, name)
                if os.path.isdir(fullname):
                    # add dir to migration session dirs list
                    # Migration session will be initialized further
                    self.__session_ids.append(name)

    def getSessionListObject(self):
        self.__get_session_ids()
        session_list_object = SessionList()
        for session_id in self.__session_ids:
            session_object = self.getSessionObject(session_id)
            if session_object is not None:
                session_list_object.add_session( session_object )
        return session_list_object

    def getSessionObject(self, session):
        migration_session = self.__get_session(session)
        if migration_session is None:
            _logger.warning("Could not initialize migration session from session_id %s. Session passed." % session)
            return None
        session_object = Session()
        # Set required attribute - always defined for migration session
        session_object.set_session_id(migration_session.getSessionId())
        start_time = migration_session.getSavedTime('start_time')
        start_time_formatted_str = time.strftime('%Y-%b-%d %H:%M:%S',start_time)
        session_object.set_creation_date(start_time_formatted_str)
        
        # Setting None is allowed
        session_object.set_steps_passed(migration_session.getStepsPassed())
        
        # Setting None is allowed
        session_object.set_additional_info(migration_session.getAdditionalInfo())
        return session_object

    def __create_unique_session(self):
        # create new dir for migration session
        # create new unique session_id
        random.seed()
        session_id = time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000))
        session_path = os.path.join( migration_handler_config.migration_session_dir(), session_id)
        while os.path.isdir(session_path):
            session_id = time.strftime('%Y%m%d%H%M%S',time.localtime()) + str(random.randint(0,1000))
            session_path = os.path.join( migration_handler_config.migration_session_dir(), session_id)
        os.makedirs(session_path)
        return session_id, session_path

    def addSession(self, agent_credentials):
        # create new dir for migration session
        # create new unique session_id
        _logger.info("Create migration session started")
        if agent_credentials is None:
            _logger.error("Could not create migration session: agent-credentials passed is None")
            raise Exception("Could not create migration session: agent-credentials passed is None")
        if not isinstance(agent_credentials, AgentCredentials):
            _logger.error("Could not create migration session: agent-credentials passed has wrong type %s" % agent_credentials.__class__.__name__)
            raise Exception("Could not create migration session: agent-credentials passed has wrong type %s" % agent_credentials.__class__.__name__)
        session_id, session_path = self.__create_unique_session()
        resp_str = cStringIO.StringIO()
        resp_encoded = EncoderFile(resp_str, "utf-8")
        resp_encoded.write('<?xml version="1.0" encoding="UTF-8"?>\n')
        agent_credentials.export(resp_encoded, 0, name_ = 'agent-credentials')
        agent_credentials_storage = os.path.join( session_path, 'agent_credentials')
        f = open(agent_credentials_storage, 'wt')
        f.write(resp_str.getvalue())
        f.close()
        session = MigrationSession.createSession(session_id)
        _logger.info("Create migration session " + str(session.getSessionId()) + " finished")
        return session

    def removeSession(self, session):
        migration_session = self.__get_session(session)
        _logger.info("Remove migration session %s started" % migration_session.getSessionId())
        dst_temporary_directory_value = migration_session.getValue('dst_temporary_directory')
        if dst_temporary_directory_value is not None:
            try:
                shutil.rmtree(dst_temporary_directory_value)
            except Exception, e:
                _logger.warning(u"Unable to remove dst-temporary-directory '%s': %s" % (dst_temporary_directory_value, unicode(e)))
        try:
            shutil.rmtree(migration_session.getSessionPath())
        except Exception, e:
            _logger.warning(u"Unable to remove session dir: %s" % unicode(e))
        _logger.info("Remove migration session finished")

